require "Window"
require "CSIsLib" 
-----------------------------------------------------------------------------------------------
-- AutoQuest Module Definition
-----------------------------------------------------------------------------------------------
local AutoQuest = {} 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local sVersion = "1.0.6"

local QuestStateEnumerationLookupTable = {}
QuestStateEnumerationLookupTable[Quest.QuestState_Unknown] = "QuestState_Unknown"
QuestStateEnumerationLookupTable[Quest.QuestState_Accepted] = "QuestState_Accepted"
QuestStateEnumerationLookupTable[Quest.QuestState_Achieved] = "QuestState_Achieved"
QuestStateEnumerationLookupTable[Quest.QuestState_Completed] = "QuestState_Completed"
QuestStateEnumerationLookupTable[Quest.QuestState_Botched] = "QuestState_Botched"
QuestStateEnumerationLookupTable[Quest.QuestState_Mentioned] = "QuestState_Mentioned"
QuestStateEnumerationLookupTable[Quest.QuestState_Abandoned] = "QuestState_Abandoned"
QuestStateEnumerationLookupTable[Quest.QuestState_Ignored] = "QuestState_Ignored"

local tIgnoredQuests = {
[8941] = true, -- 10% Group PvE Bonus!  (Housing Bilboard)
[8950] = true, -- 5% PvP Bonus!  (Housing Bilboard)
[8951] = true, -- 5% Questing and Hunting Bonus!  (Housing Bilboard)

[8949] = true, -- 5% Group PvE Bonus!  (Housing Bilboard)
[8942] = true, -- 10% PvP Bonus!  (Housing Bilboard)
[8948] = true, -- 5% Questing and Hunting Bonus!  (Housing Bilboard)

[8947] = true, -- 5% Group PvE Bonus!  (Housing Bilboard)
[8946] = true, -- 5% PvP Bonus!  (Housing Bilboard)
[8943] = true, -- 10% Questing and Hunting Bonus!  (Housing Bilboard)
}

local tDefaultOptions = {
	DebugOption = false,
	DisableOnAlt = true,
	DisableOnShift = false,
	ShowNewQuests = true,
	ShowCompletedQuests = false,
	DisableInCombat = true,
}

-----------------------------------------------------------------------------------------------
-- locals
-----------------------------------------------------------------------------------------------
local tOptions = {}
local tTimerDate = {}
local bQuestsLoaded = false
local tQueuedCommunicatorQuests = {}
local bIsPlayerInCombat = false
local bCombatDebug = false

-----------------------------------------------------------------------------------------------
-- Saving and Loading
-----------------------------------------------------------------------------------------------
function AutoQuest:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	local tSavedData = {}
	for key,value in pairs(tOptions) do  --copy the data table's top level
		tSavedData[key] = tOptions[key]
	end
	tSavedData[sVersion] = sVersion
	return tSavedData
end

function AutoQuest:OnRestore(eType, tSavedData)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	for key,value in pairs(tSavedData) do  --copy the data table's top level
		tOptions[key] = value
	end
end

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function AutoQuest:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here
    self.tCurrentQuests = {}
    return o
end

function AutoQuest:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- AutoQuest OnLoad
-----------------------------------------------------------------------------------------------
function AutoQuest:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("AutoQuest.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- AutoQuest OnDocLoaded
-----------------------------------------------------------------------------------------------
function AutoQuest:OnDocLoaded()

	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "AutoQuestForm", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
	    self.wndMain:Show(false, true)
	    self.wndMain:FindChild("Title"):SetText("AutoQuest Options v" .. sVersion)

		-- if the xmlDoc is no longer needed, you should set it to nil
		self.xmlDoc = nil
		
		Apollo.RegisterEventHandler("QuestStateChanged", "OnQuestStateChanged", self) --Used to find Mentioned/Achieved quests so we can call contact.
		Apollo.RegisterEventHandler("QuestObjectiveUpdated", "OnQuestObjectiveUpdated", self)  --Used to Find Achieved quests when QuestStateChanged doesn't fire.
		--Cant register the event OnDocLoaded or the addon breaks the Dialog system if we register our event before the Dialog addon.
		--Apollo.RegisterEventHandler("Dialog_ShowState", "OnDialog_ShowState", self) --Used for auto Quest Complete/Accept. Selects dialog options,.
		Apollo.CreateTimer("AutoQuestDialogUpdateTimer", 5, false)
		Apollo.RegisterTimerHandler("AutoQuestDialogUpdateTimer", "OnAutoQuestDialogUpdateTimer", self)
		
		Apollo.RegisterEventHandler("UnitEnteredCombat", "OnEnteredCombat", self)
		Apollo.RegisterSlashCommand("AutoQuest", "ShowOptions", self)
		Apollo.RegisterSlashCommand("autoquest", "ShowOptions", self)
		Apollo.RegisterEventHandler("Dialog_Close", "OnDialog_Close", self) --Used to know when to call the next contact

		Apollo.RegisterTimerHandler("AutoQuestCallNextContactTimer", "OnCallNextContactTimer", self)
		Apollo.CreateTimer("AutoQuestCallNextContactTimer", 3, true)
		Apollo.StopTimer("AutoQuestCallNextContactTimer")

		--Default Options
		for sOptionName, optionValue in pairs(tDefaultOptions) do
			if tOptions[sOptionName] == nil then 
				tOptions[sOptionName] = optionValue
			end
		end
	end
end

-----------------------------------------------------------------------------------------------
-- AutoQuest:OnQuestStateChanged
-- Not reliable for tracking when a quest completes.  
-- Final kill of a kill quest doesn't always trigger, Check OnQuestObjectiveUpdated
-- Used for checking quests that are QuestState_Mentioned to call the contact
-----------------------------------------------------------------------------------------------
function AutoQuest:OnQuestStateChanged(queQuest, eState)  --Called when we first get a quest
	AutoQuest:HandleQuest(queQuest)
end


-----------------------------------------------------------------------------------------------
-- AutoQuest:OnQuestObjectiveUpdated
-- Used for tracking quests that change to QuestState_Achieved and calls the contact if there is one
-----------------------------------------------------------------------------------------------
function AutoQuest:OnQuestObjectiveUpdated(queQuest, nObjective)
	AutoQuest:HandleQuest(queQuest)
end

-----------------------------------------------------------------------------------------------
-- AutoQuest:HandleQuest
-- Main quest handling logic.  Calls contacts for new quests and recently achieved quests.
-----------------------------------------------------------------------------------------------
function AutoQuest:HandleQuest(queQuest)
	if tOptions.DisableOnAlt and Apollo.IsAltKeyDown() then return end
	if tOptions.DisableOnShift and Apollo.IsShiftKeyDown() then return end
	if not bQuestsLoaded then return end
	if not queQuest then return end  --Shouldn't happen, but avoid errors. Users hate errors.

	local strTitle = queQuest:GetTitle()
	local eQuestState = queQuest:GetState()
	local strQuestState = QuestStateEnumerationLookupTable[eQuestState] or eQuestState
	local tContactInfo = queQuest:GetContactInfo()
	local nQuestId = queQuest:GetId()

	if eQuestState == Quest.QuestState_Accepted then --Still doing the quest
		if not (self.tCurrentQuests[nQuestId] == eQuestState) then --We didn't have the quest in our list, it is new.
			self.tCurrentQuests[nQuestId] = eQuestState
			if tOptions.ShowNewQuests then 
				local nConLevel = queQuest:GetConLevel()
				ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, "New Quest: " .. strTitle .. " [" .. nConLevel .. "]")
			end
		end
	elseif eQuestState == Quest.QuestState_Achieved then		
		if not tContactInfo or not tContactInfo.strName or string.len(tContactInfo.strName) <= 0 then
			AutoQuest:DebugPrint("Finished " .. strTitle  .. ". No remote turnin.")
		else
			if tOptions.DisableInCombat and bIsPlayerInCombat then
				AutoQuest:DebugPrint("Finished " .. strTitle  .. ". Waiting on Combat.")
				AutoQuest:QueueQuest(queQuest)
			else
				AutoQuest:DebugPrint("Finished " .. strTitle  .. ". Calling contact.")
				CommunicatorLib.CallContact(queQuest)
			end
		end
	elseif eQuestState == Quest.QuestState_Mentioned then
		self.tCurrentQuests[nQuestId] = eQuestState
		if not tContactInfo or not tContactInfo.strName or string.len(tContactInfo.strName) <= 0 then
			--Not sure that this state is even possible without a contact.  
			--QuestState_Mentioned is used to prompt quests without talking to a quest npc.
			AutoQuest:DebugPrint(strTitle .. " mentioned, but no contat info.")
		else
			if tOptions.DisableInCombat and bIsPlayerInCombat then
				AutoQuest:DebugPrint(strTitle  .. " mentioned. Waiting on Combat.")
				AutoQuest:QueueQuest(queQuest)
			else
				AutoQuest:DebugPrint(strTitle  .. " mentioned. Calling contact.")
				CommunicatorLib.CallContact(queQuest)
			end
		end
	elseif eQuestState == Quest.QuestState_Completed then --Quest turned in successfully.
		self.tCurrentQuests[nQuestId] = nil
		if tOptions.ShowCompletedQuests then 
			local nConLevel = queQuest:GetConLevel()
			ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, "Completed Quest: " .. strTitle .. " [" .. nConLevel .. "]")
		end
	elseif eQuestState == Quest.QuestState_Abandoned then
		self.tCurrentQuests[nQuestId] = nil
	elseif eQuestState == Quest.QuestState_Botched then
		self.tCurrentQuests[nQuestId] = nil
	else --Other states like botched could be handled, but I'm not sure what the desired response is.
		AutoQuest:DebugPrint("Quest:" .. strTitle .. " : " .. strQuestState .. ". Unhandled Case")
	end
end


-----------------------------------------------------------------------------------------------
-- AutoQuest:OnAutoQuestDialogUpdateTimer
-- Delay registering the event handler until Dialog has registered its event handler so we are called after
-- Without this delay the addon randomly breaks the Dialog system depending on the order that addons are loaded.
-----------------------------------------------------------------------------------------------
function AutoQuest:OnAutoQuestDialogUpdateTimer()
	--AutoQuest:DebugPrint("Delayed Register Timer")
	Apollo.RegisterEventHandler("Dialog_ShowState", "OnDialog_ShowState", self)
	for idx, epiEpisode in pairs(QuestLib.GetAllEpisodes(false)) do
		for idx, queQuest in pairs(epiEpisode:GetAllQuests()) do
			local eQuestState = queQuest:GetState()
			local nQuestId = queQuest:GetId()
			if eQuestState == Quest.QuestState_Accepted or 
			   eQuestState == Quest.QuestState_Achieved then
				self.tCurrentQuests[nQuestId] = eQuestState
				--AutoQuest:DebugPrint(queQuest:GetTitle() .. " : " .. QuestStateEnumerationLookupTable[eQuestState])
			end
		end
	end
	bQuestsLoaded = true
	AutoQuest:DebugPrint("AutoQuest Loaded.")
end

-----------------------------------------------------------------------------------------------
-- AutoQuest:OnDialog_ShowState
-- This event fires when talking to a quest NPC.
-- This event also triggers when calling a contact.
-- Call Select on a Dialog Response object to pick that option
-- Picks the Complete or Accept option
-----------------------------------------------------------------------------------------------
function AutoQuest:OnDialog_ShowState(eState, queCurrent)
	if Apollo.GetAddon("InstantQuestAccept") then return end
	AutoQuest:DebugPrint("OnDialog_ShowState " .. eState)

	--queCurrent = tTimerDate.queCurrent
	local idQuest = 0
	if queCurrent and queCurrent:GetId() then
		idQuest = queCurrent:GetId()
	end
	local tResponseList = DialogSys.GetResponses(idQuest)
	if not tResponseList or #tResponseList == 0 then
		return
	end
	--AutoQuest:DebugPrint(#tResponseList .. " responses")

	--Building the table of responses.
	local tQuestResponses = {}
	for idx, drResponse in ipairs(tResponseList) do
		local eResponseType = drResponse:GetType()
		local nResponseQuestId = drResponse:GetQuestId()		
		--Check for multiple response complete with the same quest ID and don't pick anything so the player can select the turnin.
		if not tIgnoredQuests[nResponseQuestId ] then
			if eResponseType == DialogResponse.DialogResponseType_QuestComplete or
				eResponseType == DialogResponse.DialogResponseType_ViewQuestComplete or
				eResponseType == DialogResponse.DialogResponseType_QuestAccept or
				eResponseType == DialogResponse.DialogResponseType_ViewQuestAccept then
				if tQuestResponses[nResponseQuestId] then -- We alreayd have a quest with this ID
					tQuestResponses[nResponseQuestId][ #tQuestResponses[nResponseQuestId] +1 ] = drResponse
				else
					tQuestResponses[nResponseQuestId] = {}
					tQuestResponses[nResponseQuestId][1] = drResponse
				end
			end
		else
			AutoQuest:DebugPrint("Ignoring quest: " .. drResponse:GetText() .. ". QuestID: " .. drResponse:GetQuestId())
		end
	end
	--Selection Logic
	for nQuestId, tResponses in pairs(tQuestResponses) do
		if #tResponses == 1 then
		if tOptions.DisableOnShift and Apollo.IsShiftKeyDown() then return end
			AutoQuest:DebugPrint("Selecting: " .. tResponses[1]:GetText() .. ". QuestID: " .. nQuestId)
			if not (tOptions.DisableOnAlt and Apollo.IsAltKeyDown()) and not (tOptions.DisableOnShift and Apollo.IsShiftKeyDown()) then
				tResponses[1]:Select()
			end
		else
			AutoQuest:DebugPrint("Multiple Choices, User select.")
		end
	end
end

-----------------------------------------------------------------------------------------------
-- AutoQuest:OnEnteredCombat
-- Used for setting the combat flag so we can queue communicator calls if the option is set
-----------------------------------------------------------------------------------------------
function AutoQuest:OnEnteredCombat(unit, bInCombat)
	if bCombatDebug then return end
	if not tOptions.DisableInCombat then return end
	if unit and unit:IsThePlayer() then
		bIsPlayerInCombat = bInCombat
		AutoQuest:DebugPrint("Combat " .. tostring(bInCombat))
		if bInCombat == false then --Just left combat
			AutoQuest:CallNextContact()
		end
	end
end

-----------------------------------------------------------------------------------------------
-- AutoQuest:FakeCombat
-- Used for debugging to set the combat flag and disable the OnEnteredCombat handler
-----------------------------------------------------------------------------------------------
function AutoQuest:FakeCombat(bInCombat)
	AutoQuest:DebugPrint("Faking Combat")
	bCombatDebug = false --Turn the handler back on
	AutoQuest:OnEnteredCombat(GameLib.GetPlayerUnit(), bInCombat) --Fake an event
	bCombatDebug = bInCombat --Disable the handler if entering combat
	Event_FireGenericEvent("SendVarToRover", "AutoQuest tQueuedCommunicatorQuests", tQueuedCommunicatorQuests, 1)
end

-----------------------------------------------------------------------------------------------
-- AutoQuest:QueueQuest
-- Adds a Quest to the queue.  
-- Used for adding communicator calls to the queue when the disable in combat option is set.
-----------------------------------------------------------------------------------------------
function AutoQuest:QueueQuest(queQuest)
	if not queQuest then return end
	if not Quest.is(queQuest) then return end
	AutoQuest:DebugPrint("QueuedQuest " .. queQuest:GetTitle())
	tQueuedCommunicatorQuests[queQuest:GetId()] = queQuest
	local numQueued = 0
	for nQueuedId, queQueuedQuest in pairs(tQueuedCommunicatorQuests) do numQueued = numQueued +1 end
	AutoQuest:DebugPrint(numQueued .. " QueuedQuest")
	if bCombatDebug then
		Event_FireGenericEvent("SendVarToRover", "AutoQuest tQueuedCommunicatorQuests", tQueuedCommunicatorQuests, 1)
	end
	Apollo.StartTimer("AutoQuestCallNextContactTimer") --Turn on Our Cleanup Timer
end

-----------------------------------------------------------------------------------------------
-- AutoQuest CallNextContact
-- Calls the first Quest in the queue.
-- Cleans up entries that don't need to be called.
-----------------------------------------------------------------------------------------------
function AutoQuest:CallNextContact()
	if bCombatDebug then
		Event_FireGenericEvent("SendVarToRover", "AutoQuest tQueuedCommunicatorQuests", tQueuedCommunicatorQuests, 1)
	end
	local eDialogState = DialogSys.GetState()
	--AutoQuest:DebugPrint("Calling next Contact " .. eDialogState)
	local queContactToCall
	for nQueuedId, queQueuedQuest in pairs(tQueuedCommunicatorQuests) do
		local tContactInfo = queQueuedQuest:GetContactInfo()
		local eState = queQueuedQuest:GetState()
		local sQuestTitle = queQueuedQuest:GetTitle()
		--Only Mentioned or Achieved Quests should be in the list
		if eState == Quest.QuestState_Mentioned or eState == Quest.QuestState_Achieved then 
			if tContactInfo and tContactInfo.strName and string.len(tContactInfo.strName) > 0 then
				if not queContactToCall then
					queContactToCall = queQueuedQuest
				end
			else
				tQueuedCommunicatorQuests[nQueuedId] = nil  --No Contact Info.
				AutoQuest:DebugPrint("No contact info for " .. sQuestTitle)
			end
		else
			tQueuedCommunicatorQuests[nQueuedId] = nil --Not Mentioned or Achieved.
			AutoQuest:DebugPrint("Bad state for contact " .. sQuestTitle)
		end
	end

	if queContactToCall then
		if tOptions.DisableInCombat and bIsPlayerInCombat then
			AutoQuest:DebugPrint("CallNextContact() Called but we are in combat.")
		else
			if eDialogState ~= DialogSys.DialogState_Inactive then
				AutoQuest:DebugPrint("Dialog window still open.")
			else
				AutoQuest:DebugPrint("CallingNextContact " .. queContactToCall:GetTitle())
				CommunicatorLib.CallContact(queContactToCall)
			end
		end
		local numQueued = 0
		for nQueuedId, queQueuedQuest in pairs(tQueuedCommunicatorQuests) do numQueued = numQueued +1 end
		AutoQuest:DebugPrint(numQueued .. " QueuedQuest")
	else
		AutoQuest:DebugPrint("No queued contacts.")
		Apollo.StopTimer("AutoQuestDialogUpdateTimer") --Turn off Our Cleanup Timer
	end
end

--OnDialogClose Doesn't always fire so we need a timer to keep checking if it is a good time to turn in quests in the queue.
function AutoQuest:OnCallNextContactTimer()
	AutoQuest:CallNextContact()
end

function AutoQuest:OnDialog_Close()
	AutoQuest:DebugPrint("OnDialog_Close " .. DialogSys.GetState())
	AutoQuest:CallNextContact()
end

-----------------------------------------------------------------------------------------------
-- AutoQuest DebugPrint
-- AutoQuest function for sending debug messages only when the debug option is set
-----------------------------------------------------------------------------------------------
function AutoQuest:DebugPrint(sMessage)
	if not tOptions.DebugOption then return end
	Print("AutoQuest: " .. sMessage)
end	

-----------------------------------------------------------------------------------------------
-- AutoQuestForm Functions
-----------------------------------------------------------------------------------------------
function AutoQuest:ShowOptions()
	self.wndMain:Show(true)
end

-- when the OK button is clicked
function AutoQuest:OnOK()
	tOptions.DebugOption			= self.wndMain:FindChild("DebugOption"):IsChecked()
	tOptions.DisableOnAlt			= self.wndMain:FindChild("AltOption"):IsChecked()
	tOptions.DisableOnShift			= self.wndMain:FindChild("ShiftOption"):IsChecked()
	tOptions.ShowNewQuests 			= self.wndMain:FindChild("ShowNewQuestsOption"):IsChecked()
	tOptions.ShowCompletedQuests 	= self.wndMain:FindChild("ShowCompletedQuestsOption"):IsChecked()
	tOptions.DisableInCombat 	= self.wndMain:FindChild("DisableInCombatOption"):IsChecked()
	self.wndMain:Close() -- hide the window
end

-- when the Cancel button is clicked
function AutoQuest:OnCancel()
	self.wndMain:Close() -- hide the window
end

function AutoQuest:OnShow(wndHandler, wndControl)
	self.wndMain:FindChild("DebugOption"):SetCheck(tOptions.DebugOption)
	self.wndMain:FindChild("AltOption"):SetCheck(tOptions.DisableOnAlt)
	self.wndMain:FindChild("ShiftOption"):SetCheck(tOptions.DisableOnShift)
	self.wndMain:FindChild("ShowNewQuestsOption"):SetCheck(tOptions.ShowNewQuests)
	self.wndMain:FindChild("ShowCompletedQuestsOption"):SetCheck(tOptions.ShowCompletedQuests)
	self.wndMain:FindChild("DisableInCombatOption"):SetCheck(tOptions.DisableInCombat)
end


-----------------------------------------------------------------------------------------------
-- AutoQuest Instance
-----------------------------------------------------------------------------------------------
local AutoQuestInst = AutoQuest:new()
AutoQuestInst:Init()
